@@ -9,7 +9,7 @@
 
 require 5.004;
 
-$DBD::ODBC::VERSION = '0.32';
+$DBD::ODBC::VERSION = '0.34';
 
 {
     package DBD::ODBC;
@@ -117,17 +117,58 @@ $DBD::ODBC::VERSION = '0.32';
 
 
     sub table_info {
-	my($dbh) = @_;		# XXX add qualification
+ 	my($dbh, $catalog, $schema, $table, $type) = @_;
+
+	if ($#_ == 1) {
+	   my $attrs = $_[1];
+	   $catalog = $attrs->{TABLE_CAT};
+	   $schema = $attrs->{TABLE_SCHEM};
+	   $table = $attrs->{TABLE_NAME};
+	   $type = $attrs->{TABLE_TYPE};
+ 	}
+
+	$catalog = "" if (!$catalog);
+	$schema = "" if (!$schema);
+	$table = "" if (!$table);
+	$type = "" if (!$type);
 
 	# create a "blank" statement handle
 	my $sth = DBI::_new_sth($dbh, { 'Statement' => "SQLTables" });
 
-	# XXX use qualification(s) (qual, schema, etc?) here...
-	DBD::ODBC::st::_tables($dbh,$sth, "")
-		or return undef;
+	DBD::ODBC::st::_tables($dbh,$sth, $catalog, $schema, $table, $type)
+	      or return undef;
 	$sth;
     }
 
+    sub primary_key_info {
+       my ($dbh, $catalog, $schema, $table ) = @_;
+ 
+       # create a "blank" statement handle
+       my $sth = DBI::_new_sth($dbh, { 'Statement' => "SQLPrimaryKeys" });
+ 
+       $catalog = "" if (!$catalog);
+       $schema = "" if (!$schema);
+       $table = "" if (!$table);
+       DBD::ODBC::st::_primary_keys($dbh,$sth, $catalog, $schema, $table )
+	     or return undef;
+       $sth;
+    }
+
+    sub foreign_key_info {
+       my ($dbh, $pkcatalog, $pkschema, $pktable, $fkcatalog, $fkschema, $fktable ) = @_;
+ 
+       # create a "blank" statement handle
+       my $sth = DBI::_new_sth($dbh, { 'Statement' => "SQLForeignKeys" });
+ 
+       $pkcatalog = "" if (!$pkcatalog);
+       $pkschema = "" if (!$pkschema);
+       $pktable = "" if (!$pktable);
+       $fkcatalog = "" if (!$fkcatalog);
+       $fkschema = "" if (!$fkschema);
+       $fktable = "" if (!$fktable);
+       _GetForeignKeys($dbh, $sth, $pkcatalog, $pkschema, $pktable, $fkcatalog, $fkschema, $fktable) or return undef;
+       $sth;
+    }
 
     sub ping {
 	my $dbh = shift;
@@ -165,6 +206,13 @@ $DBD::ODBC::VERSION = '0.32';
 	return 0;
     }
 
+    # New support for the next DBI which will have a get_info command.
+    # leaving support for ->func(xxx, GetInfo) (above) for a period of time
+    # to support older applications which used this.
+    sub get_info {
+	my ($dbh, $item) = @_;
+	return _GetInfo($dbh, $item);
+    }
 
     # Call the ODBC function SQLGetInfo
     # Args are:
@@ -287,14 +335,109 @@ See L<DBI> for more information.
 =head2 Recent Updates
 
 =over 4
+=item B<An Important note about the tests!>
 
+ Please note that some tests may fail or report they are
+ unsupported on this platform.  Notably Oracle's ODBC driver
+ will fail the "advanced" binding tests in t/08bind2.t.
+ These tests run perfectly under SQL Server 2000. This is
+ normal and expected.  Until Oracle fixes their drivers to
+ do the right thing from an ODBC perspective, it's going to
+ be tough to fix the issue.
+   
+ Also note that some tests may be skipped, such as
+ t/09multi.t, if your driver doesn't seem to support
+ returning multiple result sets.
+   
+=item B<DBD::ODBC 0.34>
+ 
+ Further revamped tests to attempt to determine if SQLDescribeParam
+ will work to handle the binding types.  The t/08bind.t attempts
+ to determine if SQLDescribeParam is supported.  note that Oracle's
+ ODBC driver under NT doesn't work correctly when binding dates
+ using the ODBC date formatting {d } or {ts }.  So, test #3 will
+ fail in t/08bind.t
+
+ New support for primary_key_info thanks to patches by Martin Evans.
+ New support for catalog, schema, table and table_type in table_info
+ thanks to Martin Evans.  Thanks Martin for your work and your
+ continuing testing, suggestions and general support!
+
+ Support for upcoming dbi get_info.
+ 
+=item B<DBD::ODBC 0.33_3>
+
+ Revamped tests to include tests for multiple result sets.
+ The tests are ODBC driver platform specific and will be skipped
+ for drivers which do not support multiple result sets.
+ 
+=item B<DBD::ODBC 0.33_2>
+
+ Finally tested new binding techniques with SQL Server 2000,
+ but there is a nice little bug in their MDAC and ODBC
+ drivers according to the knowledge base article # Q273813, titled
+
+   "FIX: "Incorrect Syntax near the Keyword 'by' "
+   Error Message with Column Names of "C", "CA" or "CAS" (Q273813)
+
+ DBD::ODBC now does not name any of the columns A, B, C, or D
+ they are now COL_A, COL_B, COL_C, COL_D.
+
+ *** NOTE: *** I AM STRONGLY CONSIDERING MAKING THE NEW
+ BINDING the default for future versions.  I do not believe
+ it will break much existing code (if any) as anyone binding
+ to non VARCHAR (without the ODBC driver doing a good conversion
+ from the VARCHAR) will have a problem.  It may be subtle, however,
+ since much code will work, but say, binding dates may not with
+ some drivers.
+   
+ Please comment soon...
+   
+=item B<DBD::ODBC 0.33_1>
+
+*** WARNING: ***
+ 
+ Changes to the binding code to allow the use of SQLDescribeParam
+ to determine if the type of column being bound.  This is
+ experimental and activated by setting
+ 
+  $dbh->{odbc_default_bind_type} = 0; # before creating the query...
+
+ Currently the default value of odbc_default_bind_type = SQL_VARCHAR
+ which mimicks the current behavior.  If you set
+ odbc_default_bind_type to 0, then SQLDescribeParam will be
+ called to determine the columen type.  Not ALL databases
+ handle this correctly.  For example, Oracle returns
+ SQL_VARCHAR for all types and attempts to convert to the
+ correct type for us.  However, if you use the ODBC escaped
+ date/time format such as: {ts '1998-05-13 00:01:00'} then
+ Oracle complains.  If you bind this with a SQL_TIMESTAMP type,
+ however, Oracle's ODBC driver will parse the time/date correctly.
+ Use at your own risk!
+
+ Fix to dbdimp.c to allow quoted identifiers to begin/end
+ with either " or '.
+ The following will not be treated as if they have a bind placeholder:
+   "isEstimated?"
+   '01-JAN-1987 00:00:00'
+   'Does anyone insert a ?'
+
+				    
+=item B<DBD::ODBC 0.32>
+
+ More SAP patches to Makfile.PL to eliminate the call to Data Sources
+
+ A patch to the test (for SAP and potentially others), to allow
+ fallback to SQL_TYPE_DATE in the tests
+ 
 =item B<DBD::ODBC 0.31>
+
  Added SAP patches to build directly against SAP driver instead of
  driver manager thanks to Flemming Frandsen (thanks!)
 
- Added support to fix ping for Oracle8.  May break other databases, so please report
- this as soon as possible.  The downside is that we need to actually execute the
- dummy query.
+ Added support to fix ping for Oracle8.  May break other databases,
+ so please report this as soon as possible.  The downside is that
+ we need to actually execute the dummy query.
  
 
 =item B<DBD::ODBC 0.30>
@@ -311,9 +454,10 @@ See L<DBI> for more information.
  the Driver upon connect.  This internal capture of the version is
  a read-only attributed and is used during array binding of parameters.
  
- Added odbc_ignore_named_placeholders attribute to facilicate creating triggers
- within SAP and Oracle, to name two. The syntax in these DBs is to allow use of
- :old and :new to access column values before and after updates.  Example:
+ Added odbc_ignore_named_placeholders attribute to facilicate
+ creating triggers within SAPDB and Oracle, to name two. The
+ syntax in these DBs is to allow use of :old and :new to
+ access column values before and after updates.  Example:
 
  $dbh->{odbc_ignore_named_placeholders} = 1; # set it for all future statements
 					  # ignores :foo, :new, etc, but not :1 or ?
